from datetime import datetime
from unittest import TestCase

import pytz
from fred_api.registry.common_types_pb2 import PlaceAddress, StateFlag
from fred_api.registry.contact.contact_info_types_pb2 import ContactAddress
from fred_api.registry.contact.service_search_contact_grpc_pb2 import SearchContactHistoryReply, SearchContactReply
from fred_api.registry.object_events_pb2 import ObjectEvents as ObjectEventsMessage

from regal.common import Address, FuzzyValue, ObjectEvent, ObjectEvents, RegistryDecoder
from regal.object import ObjectStateFlag


class AddressTest(TestCase):
    def test_str(self):
        data = (
            (Address(street=["Deck 5"], city="Red Dwarf"), "Deck 5, Red Dwarf"),
            (Address(street=["Deck 5", "Bulkhead 3"], city="Red Dwarf"), "Deck 5, Bulkhead 3, Red Dwarf"),
            (
                Address(
                    company="Jupiter Mining Corp.",
                    street=["Deck 5"],
                    city="Red Dwarf",
                    state_or_province="Quadrant Alpha",
                    postal_code="12345",
                    country_code="JMC",
                ),
                "Jupiter Mining Corp., Deck 5, Red Dwarf, Quadrant Alpha, 12345, JMC",
            ),
        )

        for address, string in data:
            with self.subTest(address=address, string=string):
                self.assertEqual(str(address), string)


class RegistryDecoderTest(TestCase):
    datetime = datetime(2020, 1, 1, tzinfo=pytz.utc)

    def setUp(self):
        self.decoder = RegistryDecoder()
        self.address = Address(
            street=["Floor 42"],
            city="Red Dwarf",
            state_or_province="JMC",
            postal_code="11111",
            country_code="XY",
        )

    def test_decode_fuzzy_value(self):
        self.assertEqual(
            self.decoder.decode(SearchContactReply.Data.FuzzyValue(lower_estimate=10, upper_estimate=100)),
            FuzzyValue(lower_estimate=10, upper_estimate=100),
        )
        self.assertEqual(
            self.decoder.decode(SearchContactHistoryReply.Data.FuzzyValue(lower_estimate=10, upper_estimate=100)),
            FuzzyValue(lower_estimate=10, upper_estimate=100),
        )
        self.assertEqual(
            self.decoder.decode(SearchContactReply.Data.FuzzyValue()), FuzzyValue(lower_estimate=0, upper_estimate=0)
        )

    def test_decode_place_address(self):
        address = PlaceAddress()
        address.street[:] = self.address.street
        address.city = self.address.city
        address.state_or_province = self.address.state_or_province
        address.postal_code.value = self.address.postal_code
        address.country_code.value = self.address.country_code

        self.assertEqual(self.decoder.decode(address), self.address)

    def test_decode_contact_address(self):
        self.address.company = "Jupiter Mining Company"

        address = ContactAddress()
        address.company = self.address.company
        address.street[:] = self.address.street
        address.city = self.address.city
        address.state_or_province = self.address.state_or_province
        address.postal_code.value = self.address.postal_code
        address.country_code.value = self.address.country_code

        self.assertEqual(self.decoder.decode(address), self.address)

    def test_decode_object_event(self):
        self.assertEqual(
            self.decoder.decode(self._make_event("REG-JMC", self.datetime)),
            ObjectEvent(registrar_handle="REG-JMC", timestamp=self.datetime),
        )

    def test_decode_object_events(self):
        events = ObjectEventsMessage()
        events.registered.CopyFrom(self._make_event("REG-JMC", self.datetime))
        events.transferred.CopyFrom(self._make_event("REG-JMC", self.datetime))
        events.updated.CopyFrom(self._make_event("REG-JMC", self.datetime))
        event = ObjectEvent(registrar_handle="REG-JMC", timestamp=self.datetime)

        self.assertEqual(
            self.decoder.decode(events),
            ObjectEvents(
                registered=event,
                transferred=event,
                updated=event,
                unregistered=None,
            ),
        )

    @staticmethod
    def _make_event(handle, datetime):
        event = ObjectEventsMessage.EventData()
        event.registrar_handle.value = handle
        event.timestamp.FromDatetime(datetime)
        return event

    def test_decode_traits(self):
        data = (
            # Traits, ObjectStateFlag
            (StateFlag.Traits(), ObjectStateFlag(manual=True, internal=False)),
            (
                StateFlag.Traits(how_to_set=StateFlag.Manipulation.Enum.manual),
                ObjectStateFlag(manual=True, internal=False),
            ),
            (
                StateFlag.Traits(how_to_set=StateFlag.Manipulation.Enum.automatic),
                ObjectStateFlag(manual=False, internal=False),
            ),
            (
                StateFlag.Traits(visibility=StateFlag.Visibility.Enum.external),
                ObjectStateFlag(manual=True, internal=False),
            ),
            (
                StateFlag.Traits(visibility=StateFlag.Visibility.Enum.internal),
                ObjectStateFlag(manual=True, internal=True),
            ),
        )
        for traits, state_flag in data:
            with self.subTest(traits=traits):
                self.assertEqual(self.decoder.decode(traits), state_flag)
